/*
 * BitTable.h - Fast reference table for filtered records.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

class DbOpen;

class BitTable
{
	friend class DbOpen;

    public:
			BitTable( DbOpen* pdb );
			~BitTable() { Free(); }

	void		Clear();
	void		Free();
	bool		Alloc( unsigned cRecords );

	bool		InView( unsigned record );
	unsigned	IndexInView( unsigned record );
	unsigned	CountInView();
	bool		FindNextRecord( unsigned* precord, bool fNext );

	bool		EnsureScanned();
	bool		RefineScan();

	bool		NewRecord( unsigned record, bool fInView );
	void		RemoveRecord( unsigned record );
	void		MoveRecord( unsigned recordFrom, unsigned recordTo );
	void		MarkRecord( unsigned record, bool fInView );
	void		RefreshRecord( unsigned record );

#ifdef DEBUG
	bool		IsInitialized() const { return m_fInitialized; }
#endif

    private:
	unsigned	RecordIndex( unsigned record ) const { return record / 32; }
	DWORD		RecordMask( unsigned record ) const { return DWORD( 1 ) << ( record % 32 ); }

	bool		FullScan();

    private:
	bool		m_fInitialized;
	unsigned	m_cAlloc;
	unsigned	m_cRecords;
	unsigned	m_cInView;
	DWORD*		m_pbits;
	DbOpen*		m_pdb;
};

