/*
 * Converter.h - Conversion tool for HandyShopper database files.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#define PURE = 0
#define IMPL

class Item;
class DbOpen;

#define DeclareOutputInterfaceMembers( IPURE ) \
	virtual int	SetFileName( TCHAR const* pszFile, bool fForceOverwrite ) IPURE; \
	virtual int	SetFriendlyNames( bool fUseFriendlyNames ) IPURE; \
	virtual int	Begin( DbOpen const* pdb ) IPURE; \
	virtual int	OutputCategory( DbOpen const* pdb, TCHAR const* pszCategory, WORD id ) IPURE; \
	virtual int	OutputPreferences( DbOpen const* pdb ) IPURE; \
	virtual int	OutputStore( DbOpen const* pdb, TCHAR const* pszStore, BYTE id ) IPURE; \
	virtual int	OutputItem( DbOpen const* pdb, Item const* pitem ) IPURE; \
	virtual int	Finished( DbOpen const* pdb ) IPURE; \

class OutputInterface : public Object
{
    public:
	DeclareOutputInterfaceMembers( PURE );
};

extern OutputInterface* NewOutputCSV( bool fSpreadsheetCompatibility );
extern OutputInterface* NewOutputText();
extern OutputInterface* NewOutputXML();

enum DateFormatType { DATEFORMAT_AUTO, DATEFORMAT_YMD, DATEFORMAT_MDY, DATEFORMAT_DMY };
void SetDateFormat( DateFormatType format );
DateFormatType GetDateFormat();

int DoImportCSV( TCHAR const* pszFrom, TCHAR const* pszTo, bool fUseStoreIds, bool fForceOverwrite );

