/*
 * db.h - Database access routines, excerpted from HS2 source code.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "viewdefs.h"
#include "dbopen.h"
#include "pdbcat.h"

#pragma pack( push, 2 )

#define CARDNO 0			// The cardNo stuff in the Palm OS never caught on; 0 is the only card.

enum
{
	newlistDefault,
	newlistShopping = newlistDefault,
	newlistShoppingPerStore,
	newlistTodo,
	newlistDated,
	newlistChecklist,
	newlistMAX
};

/*
 * ItemType.
 */

// Tags for records in list databases.
enum
{
	tagEND,
	tagDescription,
	tagNote,
	tagFlags,
	tagQuantity,
	tagPrice,
	tagPriority,
	tagDate,
	tagAlarm,
	tagUnits,
	tagMasterStoreList,
	tagStores,
	tagAisle,
	tagEntryOrder,
	tagPerStore,
	tagCustomText,
	tagAlarmMidi,
	tagIcon,
	tagUnitsEx,
	tagAisleEx,
	tagPerStoreEx,
	tagCategoryList,
	tagCategory,
	tagAutoOrder,
	tagMAX
};

// Tags for records in the prefs database.
enum
{
	tagX_RESERVED,		// Reserved.
	tagX_DBName,		// This record is a reference to a database (format: tag, string, null terminator).
	tagX_Units_Old,		// This record contains the list of user-defined units (old format, 4 bytes per unit).
	tagX_Units,		// This record contains the list of user-defined units (new format, 16 bytes per unit).
};

#define maskEND			0x00000000
#define maskMasterStoreList	0x00000000
#define maskCategoryList	0x00000000
#define maskDescription		0x00000001
#define maskFlags		0x00000002
#define maskQuantity		0x00000004
#define maskPrice		0x00000008
#define maskPriority		0x00000010
#define maskDate		0x00000020
#define maskAlarm		0x00000040
#define maskStores		0x00000080
#define maskAisle		0x00000100
#define maskEntryOrder		0x00000200
#define maskPerStore		0x00000400
#define maskCustomText		0x00000800
#define maskIcon		0x00001000
// unused			0x00002000
#define maskCATEGORY		0x00004000
//#define maskEXTENDED		0x00008000	// Indicates extended bits are in Ftr.
#define maskNote		0x00010000
//#define maskHasNote		0x00020000
#define maskUnits		0x00040000
#define maskAlarmMidi		0x00080000
#define maskAutoOrder		0x00100000
// unused			0x00200000
// unused			...
// unused			0x80000000
#define maskKNOWN		0x001d5fff	// UPDATE THIS WHEN ADDING BITS!
//#define maskHIGHWORD		0xffff0000	// Check if the mask includes ext bits.
#define maskALL			0xffff7fff	// All "real" mask bits.

/*
 * DatabasePrefs.
 */

enum { viewAll, viewNeed, viewNotNeeded, viewAllNotPurchased, viewMAX };
enum { sortNone, sortPriority, sortDate, sortAisle, sortCategory, sortDescription, sortQuantity, sortUnits, sortPrice, sortEntryOrder, sortCustomText, sortAlarm, sortIcon, sortMAX };
typedef UInt16 SortType;

#define cchMaxCurrencyHs2	( 7 )		// 6 chars, plus null terminator.

struct SortOrderInfo
{
    public:
	SortType*		addr_m_sort() { return m_sort; }
	int			dim_m_sort() { return DimensionOf( m_sort ); }
	SortType		get_m_sort( Int16 i ) const { return MacWord( m_sort[i] ); }
	void			set_m_sort( Int16 i, SortType s ) { m_sort[i] = MacWord( s ); }
    public:
	Boolean			m_fCompletedAtBottom;
	Boolean			m_rgfDescend[NUM_SORT_FIELDS];
    private:
	SortType		m_sort[NUM_SORT_FIELDS];
};
CASSERT( sizeof( SortOrderInfo ) == 10 );

#define dpfStampToday		0x0001		// On new items, set Date to Today.
#define dpfPerStorePrices	0x0002		// Separate prices/aisles per store.
#define dpfNoCompDate		0x0004		// DON'T record completion date when items are checked off.
#define dpfNoAutoQty		0x0008		// DON'T auto-reset quantities.
#define dpfAutoDelete		0x0010		// On new items, set Auto-delete to TRUE.
#define dpfDefTaxable		0x0020		// On new items, set Taxable to TRUE.
#define dpfTwoTaxes		0x0040		// Enable 2 taxes.
#define dpfTotalCurCat		0x0100		// Total command should default to the current category instead of all categories.
#define dpfSortDisabled		0x0200		// Don't sort.
#define dpfDecimals		0x0400		// Show prices with 2 decimal places (false = round to whole number).
#define dpfResetCounts		0x0800		// Reset counts next time this database is opened in FRM_MAIN.
#define dpfSeparateSorts	0x1000		// Use separate sort orders for Need and All views.
#define dpfForceResort		0x2000		// Forces resort (used by Hs2convert).

enum
{
	//$ WARNING: (chrisant) Must be unique across all builds of HS2!
	icolcxCustom		= 0,
	icolcxDate		= 1,
	icolcxCategory		= 3,
	icolcxQuantity		= 5,
	icolcxUnits		= 6,
	icolcxAisle		= 7,
	icolcxPrice		= 8,
	icolcxUnitPrice		= 9,
};

#define TAX_USE_DEFAULT		( 0xffff )

struct DatabasePrefs14
{
    public:
	IMPL_WORD( columns );
	IMPL_OR_WORD( columns );
	IMPL_AND_WORD( columns );
	IMPL_OR_WORD( flags );
    private:
	UInt16			columns;	// Which columns to display.
	UInt16			wTaxRate;	// Tax rate (8.6% == 860).
	UInt16			flags;
    public:
	UInt8			idStore;	// Id of current store.
	UInt8			reserved1;	// ???
	SortType		sort[3];	// Sort order.
    public:
	Boolean			rgfDescend[3];
    private:
	UInt16			category;	// Category.
	UInt16			view;		// View mode (viewNeed, etc).
	UInt16			wExportFlags;
    public:
	Boolean			fHideCompleted;
	Boolean			freserved1 : 1;
	Boolean			freserved2 : 1;
	Boolean			freserved3 : 1;
	Boolean			freserved4 : 1;
	Boolean			freserved5 : 1;
	Boolean			freserved6 : 1;
	Boolean			fSortCompletedAtBottomAll : 1;
	Boolean			fSortCompletedAtBottom : 1; // MUST GO LAST IN THIS 8-BITFIELD, FOR COMPATIBILITY WITH PREVIOUS BUILDS!
private:
	UInt16			cBoldAll;	// Needed items in "All Stores".
	UInt16			cBoldNo;	// Needed items in "No Stores".

	// 0613 and later.
	UInt32			ulEntryOrder;
    public:
	Boolean			fCouponsOnly;

	// Until db version 14, these were reserved/unused.
	Boolean			rgfDescendAll[3];
	SortType		sortAll[3];

	// Db version 12.
	Char			szCurrency[cchMaxCurrencyHs2];
    private:
	UInt16			wTaxRate2;
    public:
	UInt8			prioDefault;
	signed char		cMaxLinesPerItem;
	UInt8			rgColWidth[16];
};
CASSERT( sizeof( DatabasePrefs14 ) == 72 );

struct AlarmCache
{
    public:
	IMPL_DWORD( m_seconds );
	IMPL_WORD( m_icon );

    private:
	UInt32			m_seconds;	// Alarm time in seconds.

    private:
	UInt32			m_reserved1[8];	// ...future expansion.
	UInt16			m_reserved2;	// ...future expansion.

    private:
	UInt16			m_icon;		// Icon for the database (+1, so 0 means no icon).
};

struct DatabasePrefs
{
    public:
	IMPL_WORD( cBoldAll )
	IMPL_WORD( cBoldNo )
	IMPL_WORD( flags )
	IMPL_WORD( columns )
	IMPL_OR_WORD( columns )
	IMPL_AND_WORD( columns )
	IMPL_DWORD( ulEntryOrder )
	IMPL_OR_WORD( flags )
	IMPL_AND_WORD( flags )
	IMPL_WORD( wExportFlags )
	IMPL_WORD( view )
	IMPL_WORD( wTaxRate )
	IMPL_WORD( wTaxRate2 )
	IMPL_WORD( category )
	IMPL_ZERO( reserved2 )
	IMPL_ZERO( reserved5 )
	IMPL_ZERO( reserved6 )
	IMPL_ZERO( rgCategoryIcons_deprecated )
	UInt16 get_CategoryIcon_deprecated( UInt8 index ) const { return MacWord( rgCategoryIcons_deprecated[index] ); }
	void set_CategoryIcon_deprecated( UInt8 index, UInt16 icon ) { rgCategoryIcons_deprecated[index] = MacWord( icon ); }
	void SetMultipleCategories( MultipleCategories const* pmulti );
	Boolean GetMultipleCategories( MultipleCategories* pmulti ) const;
	void SetMultipleStores( MultipleStores const* pmulti );
	Boolean GetMultipleStores( MultipleStores* pmulti ) const;
    private:
	UInt16			columns;	// Which columns to display.
	UInt16			wTaxRate;	// Tax rate (8.6% == 860).
	UInt16			flags;
    public:
	UInt8			idStore;	// Id of current store.
	Boolean			fCustomCategories; // The database is using custom categories rather than the Palm OS categories.
    private:
	UInt16			reserved2[2];	// Unused.
	UInt16			grfMultipleCategories_Old;
    public:
	RGBColorType		rgbRowColor;	// Row hilite color.
    private:
	UInt16			category;	// Category.
	UInt16			view;		// View mode (viewNeed, etc).
	UInt16			wExportFlags;
    public:
	Boolean			fHideCompleted;
	UInt8			fontPlusOne;	// Font for table (plus 1, so 0 means 'not set').
    private:
	UInt16			cBoldAll;	// Needed items in "All Stores".
	UInt16			cBoldNo;	// Needed items in "No Stores".

	// 0613 and later.
	UInt32			ulEntryOrder;
    public:
	Boolean			fCouponsOnly;

	// Until db version 14, these were reserved/unused.
	Boolean			reserved5[3];	// Unused.
	UInt16			reserved6[3];	// Unused.

	// Db version 12.
	Char			szCurrency[cchMaxCurrencyHs2];
    private:
	UInt16			wTaxRate2;
    public:
	UInt8			prioDefault;
	signed char		cMaxLinesPerItem;
	UInt8			rgColWidth[16];

	// Db version 15.
	SortOrderInfo		rgOrder[NUM_SORT_ORDERS]; // Sort info for All/Need views.

	// Db version 25.
    private:
	UInt16			rgCategoryIcons_deprecated[dmRecNumCategories];

	// Db version 29.
    public:
	Boolean			m_fMultipleCategories;
	Boolean			m_fMultipleCategories_Unfiled;
	UInt8			m_rggrfMultipleCategories[(MAX_CATEGORIES + 7) / 8];
	Boolean			m_fMultipleStores;
	Boolean			m_fMultipleStores_Unfiled;
	UInt8			m_rggrfMultipleStores[(MAX_STORES + 7) / 8];

	// Db version 30.
	Boolean			m_rgfAutoOrder[NUM_SORT_ORDERS];
	Boolean			m_fLearning;
	UInt8			reserved7;
};
#define DBVersion		30
CASSERT( sizeof( DatabasePrefs ) == 158 );

// Ensure things aren't moving around in unexpected ways.
CASSERT( OffsetOf0( DatabasePrefs14, fHideCompleted ) == OffsetOf0( DatabasePrefs, fHideCompleted ) );
CASSERT( OffsetOf0( DatabasePrefs14, szCurrency ) == OffsetOf0( DatabasePrefs, szCurrency ) );
CASSERT( 72 == OffsetOf0( DatabasePrefs, rgOrder ) );

/*
 * Important old database formats.
 */

#define DBVersion_nullterminatedstrings		11	// Null terminated persisted strings were introduced in #11.
#define DBVersion_movedtodatabaseprefs		12	// Several app prefs got moved to database prefs (taxrate, taxrate2, szcurrency, deftaxable, priodefault).
#define DBVersion_separatesorts			14	// Separate sort orders for All/Need views.
#define DBVersion_rationalsortinfo		15	// Rationalized the sort info for better code/data model.
//#						16	// Force update bold counts.
//#						17	// Introduced dpfForceResort so HS2Convert doesn't have to worry about sorting.
#define DbVersion_rowcolor			18	// Added configurable row hilite color.
//#						19	// Drag resizable columns.
//#						20
#define DbVersion_alarmcache			21	// Cache for performance when updating the Alarm Manager after an alarm has been triggered.
//#						22	// Units field expanded to 15 characters.
//#						23	// Force update bold counts.
//#						24	// Force update bold counts.
#define DbVersion_categoryicons			25	// Introduce default icons for categories.
#define DbVersion_customaisles			26	// Freeform text for aisles.
//#						27	// Force update bold counts.
#define DbVersion_customcategories		28	// When activated, custom categories enable a database to have up to 100 categories.
#define DbVersion_persistmultiples		29	// Persist multiple category/store filters.
#define DbVersion_autoorder			30	// Deferred items, plus can automatically learn the order you shop for items; stores order information in a new per-store field.

/*
 * Recalc the item counts when opening any db version lower than this number.
 */

#define DBVersion_RECALCITEMCOUNTS		30

/*
 * ThisAppInfoType.
 */

struct ThisAppInfoType20
{
	// Category information is always placed in the beginning where the
	// PalmOS category.c code expects it.
	AppInfoType		appinfo;	// Category info.
	UInt8			reserved1;	// From the compiler word aligning things.
	UInt16			reserved2;
	// Settings/preferences.
	DatabasePrefs		prefs;
	// Padding.
	UInt16			dead1;
	UInt16			dead2;
	UInt16			dead3;
	UInt16			dead4;
	UInt16			reserved3;
	UInt16			reserved4;
	UInt16			reserved5;
};
CASSERT( OffsetOf0( ThisAppInfoType20, prefs ) == 280 );

struct ThisAppInfoType
{
	// Category information is always placed in the beginning where the
	// PalmOS category.c code expects it.
	AppInfoType		appinfo;	// Category info.
	UInt8			reserved1;	// From the compiler word aligning things.
	UInt16			reserved2;
	// Alarm cache -- earliest alarm in this db; this speeds updating the
	// Alarm Manager with the next alarm when an alarm has been triggered.
	AlarmCache		alarmCache;
	// Settings/preferences.
	DatabasePrefs		prefs;
	// Padding.
	UInt16			reserved3;
	UInt16			reserved4;
	UInt16			reserved5;
};
CASSERT( OffsetOf0( ThisAppInfoType, prefs ) == 320 );
CASSERT( sizeof( ThisAppInfoType ) == 484 );

typedef ThisAppInfoType* ThisAppInfoPtr;

#pragma pack( pop )

/*
 * DbOpenHs2.
 */

enum
{
#ifdef OS_PALM
	c_cMaxUndo = 3,
#else
	c_cMaxUndo = 5,
#endif
};

struct UndoInfo
{
	Item*			m_pitem;
	UInt32			m_uid;
	unsigned		m_recordModified;
	unsigned		m_recordDeleted;
};

class MasterStoreList;
class CustomCategoryList;

class DbOpenHs2 : public DbOpen
{
	typedef DbOpen base;

	friend class SortContext;

public:
				DbOpenHs2( LPCTSTR pszFile = 0 );
				~DbOpenHs2();

	static SysError NewDatabase( LPCTSTR pszFile, LPCTSTR pszDisplayName, DbOpen** ppdb );
	static SysError OpenDatabase( LPCTSTR pszFile, DbOpen** ppdb );

	Override( base, bool, CanFlush, ( bool *pfFlushSupported ) );
	Override( base, SysError, Flush, () );
	Override( base, void, Close, () );

	OverrideConst( base, LPCTSTR, Name, () );
	OverrideConst( base, bool, HasPrefix, () );
	OverrideConst( base, LPCTSTR, FileName, () );
	OverrideConst( base, unsigned, Count, () );

	Override( base, void, OmitPrefix, () );
	Override( base, SysError, SetName, ( LPCTSTR pszName ) );

	Override( base, bool, SetView, ( WORD view ) );
	Override( base, void, SetCategory, ( WORD category ) );
	Override( base, void, SetStore, ( BYTE idStore ) );
	Override( base, void, SetHideChecked, ( bool fHideChecked ) );
	Override( base, void, SetCouponsOnly, ( bool fCouponsOnly ) );
	OverrideConst( base, WORD, GetView, () );
	OverrideConst( base, WORD, GetCategory, () );
	OverrideConst( base, BYTE, GetStore, () );
	OverrideConst( base, bool, GetHideChecked, () );
	OverrideConst( base, bool, GetCouponsOnly, () );
	OverrideConst( base, FilterReason, IsRecordFiltered, ( unsigned record ) );

	Override( base, void, SetMultipleCategories, ( MultipleCategories const* pmulti ) );
	OverrideConst( base, bool, GetMultipleCategories, ( MultipleCategories* pmulti ) );

	Override( base, void, SetMultipleStores, ( MultipleStores const* pmulti ) );
	OverrideConst( base, bool, GetMultipleStores, ( MultipleStores* pmulti ) );

	Override( base, SysError, PackRecord, ( unsigned record, Item** ppitem, WORD wFlags, WORD* pces ) );
	OverrideConst( base, Item*, UnPackRecord, ( unsigned record ) );
	Override( base, SysError, DeleteRecord, ( unsigned record, bool fCanUndo ) );
	OverrideConst( base, int, GetChangeCounter, () );
	OverrideConst( base, unsigned, GetRecordsMaximum, () );
	OverrideConst( base, bool, GetItemQuickInfo, ( unsigned record, ItemQuickInfo* pqi ) );

	OverrideConst( base, SysError, GetRawRecordBytes, ( unsigned record, BYTE** ppbOut, unsigned* pcb ) );

	OverrideConst( base, bool, GetRecordUniqueId, ( unsigned record, DWORD* pdwUniqueId ) );
	OverrideConst( base, bool, FindRecordByUniqueId, ( DWORD dwUniqueId, unsigned* precord ) );

	Override( base, bool, CanUndo, () );
	Override( base, SysError, Undo, ( unsigned* precord, bool* pfDidUndo ) );
	Override( base, void, ClearUndo, () );

	OverrideConst( base, bool, IsSortedByColumn, ( ColumnIdentifier column ) );
	Override( base, bool, SetSortOneOrder, ( bool fSortOneOrder ) );
	Override( base, bool, SetSortOrderInfo, ( int index, const DbSortOrder* pso ) );
	OverrideConst( base, bool, GetSortOneOrder, () );
	OverrideConst( base, const DbSortOrder*, GetSortOrderInfo, ( int index ) );
	Override( base, SysError, SortRecord, ( unsigned recordFrom, unsigned* precordTo ) );
	Override( base, SysError, SortDatabase, () );
	Override( base, bool, SetCanLearnOrder, ( int index, bool f ) );
	OverrideConst( base, bool, CanLearnOrder, ( int index ) );
	OverrideConst( base, bool, IsLearningOrder, () );
	Override( base, bool, ActivateLearningOrder, ( bool fActivate ) );

	Override( base, void, SetColumnWidth, ( ColumnIdentifier column, int cx ) );
	OverrideConst( base, int, GetColumnWidth, ( ColumnIdentifier column ) );

	OverrideConst( base, unsigned, GetCategoriesCount, () );
	OverrideConst( base, unsigned, GetCategoriesMaximum, () );
	OverrideConst( base, LPCTSTR, GetCategoryNameById, ( WORD category ) );
	OverrideConst( base, LPCTSTR, GetCategoryNameByIndex, ( int index ) );
	OverrideConst( base, WORD, GetCategoryIdByName, ( LPCTSTR pszCategory ) );
	OverrideConst( base, WORD, GetCategoryIdByIndex, ( int index ) );
	OverrideConst( base, int, GetCategoryIndexById, ( WORD category ) );
	OverrideConst( base, WORD, GetCategoryIcon, ( WORD category ) );

	Override( base, bool, CategoryExistsById, ( WORD category ) );
	Override( base, bool, CategoryHasRecords, ( WORD category ) );
	Override( base, bool, CategoryNew, ( LPCTSTR pszCategory, WORD icon ) );
	Override( base, bool, CategoryMerge, ( WORD oldCategory, WORD newCategory ) );
	Override( base, bool, CategoryRename, ( WORD category, LPCTSTR pszCategory ) );
	Override( base, bool, CategoryDelete, ( WORD category ) );
	Override( base, bool, SetCategoryIcon, ( WORD category, WORD icon ) );

	OverrideConst( base, unsigned, GetStoresCount, () );
	OverrideConst( base, unsigned, GetStoresMaximum, () );
	OverrideConst( base, LPCTSTR, GetStoreNameById, ( BYTE idStore ) );
	OverrideConst( base, LPCTSTR, GetStoreNameByIndex, ( int index ) );
	OverrideConst( base, BYTE, GetStoreIdByName, ( LPCTSTR pszStore ) );
	OverrideConst( base, BYTE, GetStoreIdByIndex, ( int index ) );
	OverrideConst( base, int, GetStoreIndexById, ( BYTE idStore ) );

	OverrideConst( base, unsigned, GetStoreNeededItemsCount, ( BYTE idStore ) );

	Override( base, bool, StoreExistsById, ( BYTE idStore ) );
	Override( base, bool, StoreHasRecords, ( BYTE idStore ) );
	Override( base, bool, StoreNew, ( LPCTSTR pszStore ) );
	Override( base, bool, StoreNewWithId, ( LPCTSTR pszStore, BYTE idStore ) );
	Override( base, bool, StoreMerge, ( BYTE oldStore, BYTE newStore ) );
	Override( base, bool, StoreRename, ( BYTE idStore, LPCTSTR pszStore ) );
	Override( base, bool, StoreDelete, ( BYTE idStore ) );

	Override( base, void, ResetStoreItemCounts, ( bool fProgress ) );

	Override( base, void, SetRoundToWholeNumbers, ( bool fRoundWholeNumbers ) );
	Override( base, void, SetCurrency, ( LPCTSTR pszCurrency ) );
	Override( base, void, SetTaxRate, ( WORD wTaxRate ) );
	Override( base, void, SetTaxRate2, ( WORD wTaxRate2 ) );
	Override( base, void, SetTwoTaxes, ( bool fTwoTaxes ) );
	Override( base, void, SetPerStorePrices, ( bool fPerStorePrices ) );
	Override( base, void, SetAutoQuantity, ( bool fAutoQuantity ) );
	Override( base, void, SetAutoDelete, ( bool fAutoDelete ) );
	Override( base, void, SetRecordCreationDate, ( bool f ) );
	Override( base, void, SetRecordCompletionDate, ( bool f ) );
	Override( base, void, SetTaxable, ( bool f ) );
	Override( base, void, SetDefaultPriority, ( BYTE prio ) );
	Override( base, void, SetRowBackgroundColor, ( COLORREF crBackground ) );
	Override( base, void, SetFont, ( FontId id ) );
	Override( base, void, SetMaxLinesPerItem, ( int cLines ) );
	Override( base, void, SetIcon, ( WORD icon ) );
	OverrideConst( base, bool, GetRoundToWholeNumbers, () );
	OverrideConst( base, LPCTSTR, GetCurrency, () );
	OverrideConst( base, WORD, GetTaxRate, () );
	OverrideConst( base, WORD, GetTaxRate2, () );
	OverrideConst( base, bool, GetTwoTaxes, () );
	OverrideConst( base, bool, GetPerStorePrices, () );
	OverrideConst( base, bool, GetAutoQuantity, () );
	OverrideConst( base, bool, GetAutoDelete, () );
	OverrideConst( base, bool, GetRecordCreationDate, () );
	OverrideConst( base, bool, GetRecordCompletionDate, () );
	OverrideConst( base, bool, GetTaxable, () );
	OverrideConst( base, BYTE, GetDefaultPriority, () );
	OverrideConst( base, COLORREF, GetRowBackgroundColor, () );
	OverrideConst( base, FontId, GetFont, () );
	OverrideConst( base, int, GetMaxLinesPerItem, () );
	OverrideConst( base, WORD, GetIcon, () );

	OverrideConst( base, DWORD, GetNextAlarm, () );

	Override( base, void, SetExportFlags, ( WORD flags ) );
	OverrideConst( base, WORD, GetExportFlags, () );

	OverrideConst( base, bool, IsCauterized, () );

#ifdef DEBUG
	Override( base, void, AssertNoRecordLocks, ( LPCTSTR pszContext ) );
	Override( base, WORD, GetDbVersion, () );
	Override( base, bool, HasRollbackFlag, () );
#endif

	BYTE			Inline_GetStore() const { return m_prefs.idStore; }
	bool			Inline_GetPerStorePrices() const { return !!( m_prefs.get_flags() & dpfPerStorePrices ); }

	// Ideally this shouldn't be exposed, but it is needed by some
	// internal helpers that aren't on the class.

	CustomCategoryList const* Categories() const { return m_pCategories; }

protected:
	Override( base, SysError, NewDbGroupTransaction, ( DbGroupTransaction** ppOut ) );

private:
	void			SetDirty();
	void			SetStoresDirty();
	void			SetCategoriesDirty();
	bool			IsDirty() const { return m_fUserDirty; }

	void			PrepareForPPC();
	bool			CategoriesConvertFromNative();

	void			RefreshMetadataCounter();
	bool			RepairDuplicateStoreEntries();

	void			GetMasterStoreList();
	void			PutMasterStoreList();

	void			GetCategoryList();
	void			PutCategoryList();
	void			MoveCategory( WORD toCategory, WORD fromCategory );

	UInt16			GetItemFlags( unsigned record, Err *perr ) const;

	UInt8			IdFromIndex( UInt8 i );
	UInt8			IdFromIndex( UInt16 i ) { return IdFromIndex( ( UInt8 )i ); }
	UInt8			IndexFromId_Slow( UInt8 id );
	StoreDef*		StoreFromIndex( UInt8 i );
	StoreDef*		StoreFromIndex( UInt16 i ) { return StoreFromIndex( ( UInt8 )i ); }

	void			FillInStore( UInt8 id, PCSZ psz, Boolean fNew );
	UInt8			HashStoreId( PCSZ psz );
	void			IncStore( Int8 n );
	void			ZapStore( UInt8 id );
	UInt8			StoreNewRaw( PCSZ psz, UInt8 id = noStore );
	void			StoreDeleteIndexEntryRaw( UInt8 index );
	void			UpdateCounts( StoreList* pstores, Int16 delta );
	void			ResetCounts( bool fProgress );

	Err			InitAppInfo();
	Err			ReadAppInfo( UInt16 flags );
	void			WriteAppInfo();

	void			IncrementChangeCounter();

	void			RefreshCategoryIndex();

	MaskType		GetUnpackSortMask();
	DirtyType		GetDirtyMask();

	bool			FAlreadySorted( unsigned record );
	void			SortStores( bool fDirty = true );
	SysError		SortDatabaseInternal( bool fDirty = true );

	bool			IsCategoryFiltered( WORD category ) const;

	void			SetupUndo( Item** ppitemUndo, unsigned record, bool fDeleted, bool fNew );

	void			SetPrefFlag( UInt16 flag, bool f );

	static int		CompareStores( MemPtr pA, MemPtr pB, Int32 other );
	static Int16		CmpFindSort( MemPtr _pA, MemPtr _pB, Int32 other );

	static DbOpenHs2*	DmNewDatabase( LPCTSTR pszFile, LPCSTR pszName );
	static DbOpenHs2*	DmOpenDatabase( LPCTSTR pszFile, SysError& e, IProgress* pProgress = 0 );

private:
	// Database info.
	DmOpenRef		m_pdb;
	DatabasePrefs		m_prefs;

	// Master store list.
	MasterStoreList*	m_pStores;
	bool			m_fStoresDirty;

	// Custom category list.
	CustomCategoryList*	m_pCategories;
	bool			m_fCategoriesDirty;

	// Flags.
	bool			m_fReadonly;
	bool			m_fOmitPrefix;

	mutable StrT		m_sName;	// Friendly name.
	mutable StrT		m_sFile;	// File name and extension (no path).
	mutable StrT		m_sCurrency;

	mutable int		m_iCat;		// Rotating buffers for.
	mutable StrT		m_rgCat[2];	// Returning category names.
	mutable int		m_iStore;	// Rotating buffers for.
	mutable StrT		m_rgStore[2];	// Returning store names.

	unsigned		m_cMetadataRecords;

	int			m_nChangeCounter;

	UInt16			m_cUndo;
	UndoInfo		m_rgUndo[c_cMaxUndo];

#ifdef OS_WINDOWS
	DWORD			m_codepage;
#endif

	bool			m_fDirty;	// Database has been altered.
	bool			m_fUserDirty;	// Database has unsaved edits that were made by the user (i.e. the database was somehow altered after being opened).
	bool			m_fPrepared;	// Database was successfully opened.
	bool			m_fMakeBackup;	// Make a backup on the next Flush().
	bool			m_fFlushing;	// Flush is currently in progress; avoid reentrancy!

	MultipleCategories	m_multiCategories;
	MultipleStores		m_multiStores;
};

/*
 * Alarms.
 */

#define DbgMsg_Alarms( x )	//DbgMsg( x )

void UpdateAlarm( DmOpenRef pdb, UInt32 alarmSeconds, bool fSet );

/*
 * SortContext.
 */

class SortContext
{
    public:
	SortContext( DbOpenHs2* popen );
	~SortContext();

    public:
	static void SetSortDatabase( DbOpenHs2* popen );
	static DbOpenHs2* GetSortDatabase();
};

/*
 * Functions.
 */

#ifdef DEBUG
void DbgResetUnpackCounter( void );
UInt16 DbgGetUnpackCounter( void );
#endif

inline void *ByteOffset( void *pv, UInt16 offset ) { return reinterpret_cast<UInt8*>( pv ) + offset; }

ItemPtr UnPackRecordPtr( MemPtr pRec, MaskType mask, ItemPtr pitem, unsigned recDebug = noRecordSelected );

Err CreateDatabase( char const* psz, UInt16 flags, DbOpenHs2* popen );
Err OpenDatabaseByID( UInt16 cardNo, LocalID lid, Boolean fWritable, DbOpenHs2* popen );
Err OpenDatabase( PCSZ psz, Boolean fWritable, DbOpenHs2* popen );

void SkipPrefix( char const** ppsz );

bool FindMetadataRecord( DmOpenRef pdb, UInt8 tagFind, unsigned* precord, MemHandle* phandle, void** ppv );
inline bool FindMetadataRecord( DmOpenRef pdb, UInt8 tagFind, unsigned* precord, MemHandle* phandle, MasterStoreList** ppStores ) { return FindMetadataRecord( pdb, tagFind, precord, phandle, ( void** )ppStores ); }
inline bool FindMetadataRecord( DmOpenRef pdb, UInt8 tagFind, unsigned* precord, MemHandle* phandle, CustomCategoryList** ppCategories ) { return FindMetadataRecord( pdb, tagFind, precord, phandle, ( void** )ppCategories ); }

