/*
 * Dbg.h - Debugging aids.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#if defined( _MSC_VER )
#pragma warning( disable: 4127 )
#endif

#define CASSERT( expression )		typedef int dummy_array__[( expression ) ? 1 : -1]
#define CASSERT_INLINE( expression )	( sizeof( char[( expression ) ? 1 : -1] ) - sizeof( char ) )
#define CASSERT_			CASSERT

#ifdef DEBUG
void EnableTracing( bool fEnable );
void EnableAsserts( bool fEnable );
bool AreAssertsEnabled();
#endif

#ifdef DEBUG
#define Assert( x )			do { if ( !( x ) ) { __DbgAssert( TEXT( #x ), TEXT( __FILE__ ), __LINE__ ); } } while ( 0 )
#define AssertSz( x, sz )		do { if ( !( x ) ) { __DbgAssert( TEXT( sz ), TEXT( __FILE__ ), __LINE__ ); } } while ( 0 )
#define Assert1( x, sz, arg1 )		do { if ( !( x ) ) { __DbgAssert( TEXT( sz ), TEXT( __FILE__ ), __LINE__, arg1 ); } } while ( 0 )
#define Assert2( x, sz, arg1, arg2 )	do { if ( !( x ) ) { __DbgAssert( TEXT( sz ), TEXT( __FILE__ ), __LINE__, arg1, arg2 ); } } while ( 0 )
#define Assert3( x, sz, arg1, arg2, arg3 ) do { if ( !( x ) ) { __DbgAssert( TEXT( sz ), TEXT( __FILE__ ), __LINE__, arg1, arg2, arg3 ); } } while ( 0 )
#define Assert4( x, sz, arg1, arg2, arg3, arg4 ) do { if ( !( x ) ) { __DbgAssert( TEXT( sz ), TEXT( __FILE__ ), __LINE__, arg1, arg2, arg3, arg4 ); } } while ( 0 )
#define Trace( args )			__DbgTrace args
#define DbgMsg( args )			__DbgMsg args
#define DbgDump( p, c )			__DbgDump( p, c )
#define IfDebug( x )			x
#else
#define Assert( expr )			syntax_safe
#define AssertSz( expr, sz )		syntax_safe
#define Assert1( expr, sz, arg1 )	syntax_safe
#define Assert2( expr, sz, arg1, arg2 )	syntax_safe
#define Assert3( expr, sz, arg1, arg2, arg3 ) syntax_safe
#define Assert4( expr, sz, arg1, arg2, arg3, arg4 ) syntax_safe
#define Trace( args )			syntax_safe
#define DbgMsg( args )			syntaxsafe
#define DbgDump( p, c )			syntaxsafe
#define IfDebug( x )
#endif

#define Panic()				AssertSz( 0, "Panic!" )
#define PanicSz( sz )			AssertSz( 0, sz )
#define Panic1( sz, arg1 )		Assert1( 0, sz, arg1 )
#define Panic2( sz, arg1, arg2 )	Assert2( 0, sz, arg1, arg2 )
#define Panic3( sz, arg1, arg2, arg3 )	Assert3( 0, sz, arg1, arg2, arg3 )
#define Panic4( sz, arg1, arg2, arg3, arg4 ) Assert4( 0, sz, arg1, arg2, arg3, arg4 )

#ifdef DEBUG
enum AssertResponse { AR_ABORT, AR_BREAK, AR_CONTINUE };
void __DbgTrace( LPCSTR pszFmt, ... );
#ifdef UNICODE
void __DbgTrace( LPCWSTR pszFmt, ... );
#endif
AssertResponse __DbgAssertInternalV( LPCTSTR pszFmt, LPCTSTR pszFile, int nLine, va_list args );
AssertResponse __DbgAssertInternal( LPCTSTR pszFmt, LPCTSTR pszFile, int nLine, ... );
void __DbgAssert( LPCTSTR pszFmt, LPCTSTR pszFile, int nLine, ... );
void __DbgMsg( int tenths, LPCSTR pszFormat, ... );
#ifdef UNICODE
void __DbgMsg( int tenths, LPCWSTR pszFormat, ... );
#endif
void __DbgDump( void const* pv, unsigned cb );
#endif

#if defined( OS_WINDOWS )
inline void __DbgAbort() { ExitProcess( UINT( -1 ) ); };	// Does not run any global destructors, etc (yay).
#define __DbgBreak DebugBreak
#elif defined( OS_CRT )
inline void __DbgAbort() { exit( -1 ); }			// Runs global destructors, etc (boo).
inline void __DbgBreak() { *reinterpret_cast<unsigned*>( 0 ) = 0; }
#else
#error __DbgBreak not implemented for this platform.
#endif

