/*
 * GlobalHS2.h - Definitions and globals from HS2.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#ifdef DECLARE_GLOBALS
#define GLOBAL( typ, var ) typ var;
#define GLOBALx( typ, var )
#define GLOBAL_( typ, var, val ) typ var = val;

#else
#define GLOBAL( typ, var ) extern typ var;
#define GLOBALx( typ, var ) extern typ var;
#define GLOBAL_( typ, var, val ) extern typ var;

#endif

/*
 * Palm OS 3.1
 */

#define	chrBullet		0x0095		// Bullet character (149).

/*
 * Types
 */

#define cchMaxDBName		( 48 )

typedef const Char* PCSZ;

/*
 * Defines
 */

#define AppType			'2Shp'
#define DBType			'List'

#define firstRecord		0		// Index of first data record.

#define noError			0		// Err is 0 when no error.

#define F( x )			( !!( x ) )
inline bool IfGlobals()		{ return true; }
#define ASSERTGLOBALS		syntax_safe
#define AssertSameSize( x, y )	Assert( SameSize( x, y ) )
#define CALLBACK_PROLOGUE
#define CALLBACK_EPILOGUE

// Shows a DbgMsg anytime CheckErr (etc) finds an error.
#define MSGONERROR_dm( fn, ln )		DbgMsg(( 30, TEXT( "%s, li %d:\nDmGetLastError() == %d" ), TEXT( fn ), ln, DmGetLastErr() ))
#define MSGONERROR_e( fn, ln, x )	DbgMsg(( 30, TEXT( "%s, li %d:\nerror %u\n%s" ), TEXT( fn ), ln, err, TEXT( x ) ))
#define MSGONERROR_( x, fn, ln )	DbgMsg(( 30, TEXT( "%s, li %d:\n%s" ), TEXT( fn ), ln, TEXT( x ) ))
#define MSGONERROR( x )			DbgMsg(( 30, TEXT( "%s" ), TEXT( x ) ))

//#define dmCheckErr( x )		do { err = ( x ); if ( err ) { MSGONERROR_( #x, __FILE__, __LINE__ ); goto LError; } } while ( false )
//#define dmCheckErrVoid( x )		do { if ( x ) { MSGONERROR_( #x, __FILE__, __LINE__ ); goto LError; } } while ( false )
//#define dmCE( x )			dmCheckErr( x )
//#define dmCEv( x )			dmCheckErrVoid( x )
#define dmCPE( p )			do { if ( !( p ) ) { err = DmGetLastErr(); MSGONERROR_dm( __FILE__, __LINE__ ); goto LError; } } while ( false )
#define dmCPEv( p )			do { if ( !( p ) ) { MSGONERROR_dm( __FILE__, __LINE__ ); goto LError; } } while ( false )

#define CheckErr( x )			do { err = ( x ); if ( err ) { MSGONERROR_e( __FILE__, __LINE__, #x ); goto LError; } } while ( false )
#define CheckErrVoid( x )		do { if ( x ) { MSGONERROR_( #x, __FILE__, __LINE__ ); goto LError; } } while ( false )
#define CE( x )				CheckErr( x )
#define CEv( x )			CheckErrVoid( x )
#define CPE( p )			do { if ( !( p ) ) { MSGONERROR_( #p, __FILE__, __LINE__ ); err = memErrNotEnoughSpace; goto LError; } } while ( false )
#define CPEv( p )			do { if ( !( p ) ) { MSGONERROR_( #p, __FILE__, __LINE__ ); goto LError; } } while ( false )

#define CheckErr_nomsg( x )		do { err = ( x ); if ( err ) { goto LError; } } while ( false )
#define CheckErrVoid_nomsg( x )		do { if ( x ) { goto LError; } } while ( false )
#define CE_nomsg( x )			CheckErr_nomsg( x )
#define CEv_nomsg( x )			CheckErrVoid_nomsg( x )
#define CPE_nomsg( p )			do { if ( !( p ) ) { err = memErrNotEnoughSpace; goto LError; } } while ( false )
#define CPEv_nomsg( p )			do { if ( !( p ) ) { goto LError; } } while ( false )

